# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

import ctypes
import gc
import os.path
import sys
import threading
import webbrowser
from typing import List, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt



# GUI imports
from PySide6 import QtGui, QtWidgets, QtCore

# Engine imports
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
import VeraGridEngine.Simulations as sim
from VeraGridEngine.enumerations import EngineType, DeviceType, SimulationTypes
from VeraGridEngine.Compilers.circuit_to_data import compile_numerical_circuit_at
from VeraGridEngine.DataStructures.numerical_circuit import NumericalCircuit

from VeraGridEngine.Compilers.circuit_to_bentayga import BENTAYGA_AVAILABLE
from VeraGridEngine.Compilers.circuit_to_newton_pa import NEWTON_PA_AVAILABLE
from VeraGridEngine.Compilers.circuit_to_gslv import GSLV_AVAILABLE
from VeraGridEngine.Compilers.circuit_to_pgm import PGM_AVAILABLE
from VeraGridEngine.Simulations.Clustering.clustering_results import ClusteringResults

import VeraGrid.Gui.gui_functions as gf
import VeraGrid.Session.synchronization_driver as syncdrv
from VeraGrid.Gui.AboutDialogue.about_dialogue import AboutDialogueGuiGUI
from VeraGrid.Gui.Analysis.AnalysisDialogue import GridAnalysisGUI
from VeraGrid.Gui.ContingencyPlanner.contingency_planner_dialogue import ContingencyPlannerGUI
from VeraGrid.Gui.CoordinatesInput.coordinates_dialogue import CoordinatesInputGUI
from VeraGrid.Gui.general_dialogues import CheckListDialogue, StartEndSelectionDialogue
from VeraGrid.Gui.messages import yes_no_question, warning_msg, info_msg, error_msg
from VeraGrid.Gui.GridGenerator.grid_generator_dialogue import GridGeneratorGUI
from VeraGrid.Gui.LoadCatalogue.catalogue_dialogue import CatalogueGUI
from VeraGrid.Gui.Main.MainWindow import Ui_mainWindow, QMainWindow
from VeraGrid.Gui.Main.object_select_window import ObjectSelectWindow
from VeraGrid.Gui.ProfilesInput.models_dialogue import ModelsInputGUI
from VeraGrid.Gui.ProfilesInput.profile_dialogue import ProfileInputGUI
from VeraGrid.Session.session import SimulationSession, GcThread
from VeraGrid.Gui.SigmaAnalysis.sigma_analysis_dialogue import SigmaAnalysisGUI
from VeraGrid.Gui.SyncDialogue.sync_dialogue import SyncDialogueWindow
from VeraGrid.Gui.TowerBuilder.LineBuilderDialogue import TowerBuilderGUI
from VeraGrid.Gui.GridReduce.grid_reduce import GridReduceDialogue
from VeraGrid.Gui.RmsModelEditor.rms_model_editor_engine import RmsModelEditorGUI
from VeraGrid.Gui.Diagrams.SchematicWidget.diagram_bus_selection_dialogue import DiagramBusSelectorDialogue
from VeraGrid.Gui.rms_events_editor_dialog import RmsEventEditor
from VeraGrid.Gui.Diagrams.generic_graphics import IS_DARK
from VeraGrid.Gui.python_console import PythonConsole
from VeraGrid.Gui.python_script_editor import PythonCodeEditor
from VeraGrid.Gui.toast_widget import ToastManager



def terminate_thread(thread):
    """
    Terminates a python thread from another thread.

    :param thread: a threading.Thread instance
    """
    if not thread.is_alive():
        return False

    exc = ctypes.py_object(SystemExit)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(thread.ident), exc)
    if res == 0:
        print("nonexistent thread id")
        return True

    elif res > 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")

    return True


def traverse_objects(name, obj, lst: list, i=0):
    """

    :param name:
    :param obj:
    :param lst:
    :param i:
    """
    lst.append((name, sys.getsizeof(obj)))
    if i < 10:
        if hasattr(obj, '__dict__'):
            for name2, obj2 in obj.__dict__.items():
                if isinstance(obj2, np.ndarray):
                    lst.append((name + "/" + name2, sys.getsizeof(obj2)))
                else:
                    if isinstance(obj2, list):
                        # list or
                        for k, obj3 in enumerate(obj2):
                            traverse_objects(name=name + "/" + name2 + '[' + str(k) + ']',
                                             obj=obj3, lst=lst, i=i + 1)
                    elif isinstance(obj2, dict):
                        # list or
                        for name3, obj3 in obj2.items():
                            traverse_objects(name=name + "/" + name2 + '[' + name3 + ']',
                                             obj=obj3, lst=lst, i=i + 1)
                    else:
                        # normal obj
                        if obj2 != obj:
                            traverse_objects(name=name + "/" + name2, obj=obj2, lst=lst, i=i + 1)


class BaseMainGui(QMainWindow):
    """
    DiagramFunctionsMain
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """
        # create main window
        QMainWindow.__init__(self, parent)
        self.ui = Ui_mainWindow()
        self.ui.setupUi(self)

        # Declare circuit
        self.circuit: MultiCircuit = MultiCircuit()

        self.lock_ui = False
        self.ui.progress_frame.setVisible(self.lock_ui)

        self.stuff_running_now: List[SimulationTypes] = list()

        self.session: SimulationSession = SimulationSession(name='GUI session')

        self._file_name = ''

        self.project_directory = os.path.expanduser("~")

        self.current_boundary_set: str = ""

        # toast manager
        self.toast_manager = ToastManager(parent=self, position_top=False)

        # threads ------------------------------------------------------------------------------------------------------
        self.painter = None
        self.open_file_thread_object = None
        self.save_file_thread_object = None
        self.last_file_driver = None
        self.delete_and_reduce_driver = None
        self.export_all_thread_object = None
        self.topology_reduction = None
        self.find_node_groups_driver: Union[sim.NodeGroupsDriver, None] = None
        self.file_sync_thread = syncdrv.FileSyncThread(self.circuit, None, None)

        self.task_pool = QtCore.QThreadPool()

        # simulation start end
        self.simulation_start_index: int = -1
        self.simulation_end_index: int = -1

        # Add console --------------------------------------------------------------------------------------------------
        self.console = PythonConsole(banner="VeraGrid Python Console!")
        self.ui.consoleLayout.addWidget(self.console)

        self.code_editor = PythonCodeEditor(vars_dict={

            # "app": self,
            # "np": np,
            # "pd": pd,
            # "plt": plt,
        }, )
        self.ui.codeEditorLayout.addWidget(self.code_editor)

        # window pointers ----------------------------------------------------------------------------------------------
        self.file_sync_window: Union[SyncDialogueWindow, None] = None
        self.sigma_dialogue: Union[SigmaAnalysisGUI, None] = None
        self.grid_generator_dialogue: Union[GridGeneratorGUI, None] = None
        self.catalogue_dialogue: Union[CatalogueGUI, None] = None
        self.contingency_planner_dialogue: Union[ContingencyPlannerGUI, None] = None
        self.analysis_dialogue: Union[GridAnalysisGUI, None] = None
        self.profile_input_dialogue: Union[ProfileInputGUI, None] = None
        self.models_input_dialogue: Union[ModelsInputGUI, None] = None
        self.object_select_window: Union[ObjectSelectWindow, None] = None
        self.coordinates_window: Union[CoordinatesInputGUI, None] = None
        self.about_msg_window: Union[AboutDialogueGuiGUI, None] = None
        self.tower_builder_window: Union[TowerBuilderGUI, None] = None
        self.rms_model_Editor_window: Union[RmsModelEditorGUI, None] = None
        self.rms_events_Editor_window: Union[RmsEventEditor, None] = None
        self.investment_checks_diag: Union[CheckListDialogue, None] = None
        self.new_se_dlg: Union[CheckListDialogue, None] = None
        self.contingency_checks_diag: Union[CheckListDialogue, None] = None
        self.ra_checks_diag: Union[CheckListDialogue, None] = None
        self.start_end_dialogue_window: Union[StartEndSelectionDialogue, None] = None
        self.grid_reduction_dialogue: GridReduceDialogue | None = None
        self.select_bus_dlg: DiagramBusSelectorDialogue | None = None

        # available engines --------------------------------------------------------------------------------------------
        engine_lst = [EngineType.VeraGrid]
        if GSLV_AVAILABLE:
            engine_lst.append(EngineType.GSLV)
        if NEWTON_PA_AVAILABLE:
            engine_lst.append(EngineType.NewtonPA)
        if BENTAYGA_AVAILABLE:
            engine_lst.append(EngineType.Bentayga)
        if PGM_AVAILABLE:
            engine_lst.append(EngineType.PGM)

        self.ui.engineComboBox.setModel(gf.get_list_model([x.value for x in engine_lst]))
        self.ui.engineComboBox.setCurrentIndex(0)
        self.engine_dict = {x.value: x for x in engine_lst}

        # available engines --------------------------------------------------------------------------------------------
        exchange_places = [DeviceType.AreaDevice, DeviceType.CountryDevice]
        exchange_places_mdl = gf.get_list_model([x.value for x in exchange_places])
        self.ui.fromComboBox.setModel(exchange_places_mdl)
        self.ui.fromComboBox.setCurrentIndex(0)
        self.ui.toComboBox.setModel(exchange_places_mdl)
        self.ui.toComboBox.setCurrentIndex(0)
        self.exchange_places_dict = {x.value: x for x in exchange_places}

        # dark mode detection ------------------------------------------------------------------------------------------
        self.ui.dark_mode_checkBox.setChecked(IS_DARK)

        self.calculation_inputs_to_display = None

        # --------------------------------------------------------------------------------------------------------------
        self.ui.actionClear_stuff_running_right_now.triggered.connect(self.clear_stuff_running)
        self.ui.actionAbout.triggered.connect(self.about_box)
        self.ui.actionAuto_rate_branches.triggered.connect(self.auto_rate_branches)
        self.ui.actionDetect_transformers.triggered.connect(self.detect_transformers)
        self.ui.actionLaunch_data_analysis_tool.triggered.connect(self.display_grid_analysis)
        self.ui.actionOnline_documentation.triggered.connect(self.show_online_docs)
        self.ui.actionReport_a_bug.triggered.connect(self.report_a_bug)

        self.ui.actionFix_generators_active_based_on_the_power.triggered.connect(
            self.fix_generators_active_based_on_the_power
        )
        self.ui.actionFix_loads_active_based_on_the_power.triggered.connect(self.fix_loads_active_based_on_the_power)
        self.ui.actionInitialize_contingencies.triggered.connect(self.initialize_contingencies)

        # Buttons
        self.ui.cancelButton.clicked.connect(self.set_cancel_state)

        # doubleSpinBox
        self.ui.fbase_doubleSpinBox.valueChanged.connect(self.change_circuit_base)
        self.ui.sbase_doubleSpinBox.valueChanged.connect(self.change_circuit_base)

        self.ui.grid_name_line_edit.textChanged.connect(self.change_circuit_name)

        # combo-boxes
        self.ui.fromComboBox.currentTextChanged.connect(self.update_from_to_list_views)
        self.ui.toComboBox.currentTextChanged.connect(self.update_from_to_list_views)

    def LOCK(self, val: bool = True) -> None:
        """
        Lock the interface to prevent new simulation launches
        :param val:
        :return:
        """
        self.lock_ui = val
        self.ui.progress_frame.setVisible(self.lock_ui)
        QtGui.QGuiApplication.processEvents()

    def UNLOCK(self) -> None:
        """
        Unlock the interface
        """
        if not self.any_thread_running():
            self.LOCK(False)

    @property
    def file_name(self) -> str:
        """
        Get the current file name
        :return: str
        """
        return self._file_name

    @file_name.setter
    def file_name(self, val: str):
        """
        Set the current file name
        :param val: file name
        """
        self._file_name = val
        if not isinstance(self._file_name, list):
            self.ui.file_information_label.setText(self._file_name)
        else:
            self.ui.file_information_label.setText("")

    @staticmethod
    def collect_memory() -> None:
        """
        Collect memory
        """
        for i in (0, 1, 2):
            gc.collect(generation=i)

    def get_simulation_threads(self) -> List[GcThread]:
        """
        Get all threads that has to do with simulation
        :return: list of simulation threads
        """

        all_threads = list(self.session.threads.values())

        return all_threads

    def get_process_threads(self) -> List[GcThread]:
        """
        Get all threads that has to do with processing
        :return: list of process threads
        """
        all_threads = [self.open_file_thread_object,
                       self.save_file_thread_object,
                       self.painter,
                       self.delete_and_reduce_driver,
                       self.export_all_thread_object,
                       self.find_node_groups_driver,
                       self.file_sync_thread,
                       ]
        return all_threads

    def get_all_threads(self) -> List[GcThread]:
        """
        Get all threads
        :return: list of all threads
        """
        all_threads = self.get_simulation_threads() + self.get_process_threads()
        return all_threads

    def stop_all_threads(self):
        """
        Stop all running threads
        """
        for thr in self.get_all_threads():
            if thr is not None:
                thr.quit()

        for thread in threading.enumerate():
            print(thread.name, end="")
            if "MainThread" not in thread.name:
                stat = terminate_thread(thread)
                if stat:
                    print(" killed")
                else:
                    print(" not killed")
            else:
                print(" Skipped")

        # second pass, kill main too
        for thread in threading.enumerate():
            print(thread.name, end="")
            stat = terminate_thread(thread)
            if stat:
                print(" killed")
            else:
                print(" not killed")

    def any_thread_running(self) -> bool:
        """
        Checks if any thread is running
        :return: True/False
        """
        val = False

        # this list cannot be created only once, because the None will be copied
        # instead of being a pointer to the future value like it would in a typed language
        all_threads = self.get_all_threads()

        for thr in all_threads:
            if thr is not None:
                if thr.isRunning():
                    return True
        return val

    def clear_stuff_running(self) -> None:
        """
        This clears the list of stuff running right now
        this list blocks new executions of the same threads.
        Cleaning is useful if a particular thread crashes and you want to retry.
        """
        self.stuff_running_now.clear()

    def get_all_objects_in_memory(self):
        """
        Get a list of the objects in memory
        :return:
        """
        objects = []
        # for name, obj in globals().items():
        #     objects.append([name, sys.getsizeof(obj)])

        traverse_objects('MainGUI', self, objects)

        df = pd.DataFrame(data=objects, columns=['Name', 'Size (kb)'])
        df.sort_values(by='Size (kb)', inplace=True, ascending=False)
        return df

    def expand_object_tree_nodes(self) -> None:
        """
        Expand objects' tree nodes
        """
        proxy = self.ui.dataStructuresTreeView.model()

        for row in range(proxy.rowCount()):
            index = proxy.index(row, 0)
            self.ui.dataStructuresTreeView.expand(index)

    def get_simulation_start(self) -> int:
        """
        Get the start simulation index
        """
        return self.simulation_start_index

    def get_simulation_end(self) -> int:
        """
        Get the end simulation index
        """
        return self.simulation_end_index

    def setup_sim_indices(self, st: int, en: int):
        """
        Set the simulation indices
        :param st: start index
        :param en: end index
        """
        self.simulation_start_index = st
        self.simulation_end_index = en

    def get_diagram_slider_index(self) -> Union[int, None]:
        """
        Get the diagram slider value
        :return: [None, int]
        """
        idx = self.ui.diagram_step_slider.value()
        return idx if idx > -1 else None

    def get_db_slider_index(self) -> Union[int, None]:
        """
        Get the db slider value
        :return: [None, int]
        """
        idx = self.ui.db_step_slider.value()
        return idx if idx > -1 else None

    def setup_time_sliders(self):
        """
        Setup the time sliders
        """
        for slider in [self.ui.diagram_step_slider,
                       self.ui.db_step_slider,
                       self.ui.compiled_arrays_step_slider]:
            if self.circuit.has_time_series:
                slider.setRange(-1, self.circuit.get_time_number() - 1)
            else:
                slider.setRange(-1, -1)
            slider.setValue(-1)

    def update_date_dependent_combos(self):
        """
        update the drop down menus that display dates
        """
        if self.circuit.time_profile is not None:
            mdl = gf.get_list_model([str(e) for e in self.circuit.time_profile])
            # setup profile sliders
            t = len(self.circuit.time_profile) - 1
            self.setup_sim_indices(0, t)

        else:
            mdl = QtGui.QStandardItemModel()
            self.setup_sim_indices(-1, -1)
        self.ui.vs_departure_comboBox.setModel(mdl)
        self.ui.vs_target_comboBox.setModel(mdl)
        self.setup_time_sliders()

    def update_from_to_list_views(self):
        """
        Update the exchange area, countries, etc... dependent combos
        """
        same_type = self.ui.fromComboBox.currentText() == self.ui.toComboBox.currentText()

        devs_from = self.circuit.get_elements_by_type(self.exchange_places_dict[self.ui.fromComboBox.currentText()])
        mdl1 = gf.get_list_model([str(elm) for elm in devs_from], checks=True)
        self.ui.fromListView.setModel(mdl1)

        devs_to = self.circuit.get_elements_by_type(self.exchange_places_dict[self.ui.toComboBox.currentText()])
        mdl2 = gf.get_list_model([str(elm) for elm in devs_to], checks=True)
        self.ui.toListView.setModel(mdl2)

        if len(devs_from) > 1 and same_type:
            self.ui.fromListView.model().item(0).setCheckState(QtCore.Qt.CheckState.Checked)
            self.ui.toListView.model().item(1).setCheckState(QtCore.Qt.CheckState.Checked)

    def fix_generators_active_based_on_the_power(self, ask_before=True):
        """
        set the generators active based on the active power values
        :return:
        """

        if ask_before:
            ok = yes_no_question("This action sets the generation active profile based on the active power profile "
                                 "such that ig a generator active power is zero, the active value is false",
                                 "Set generation active profile")
        else:
            ok = True

        if ok:
            self.circuit.set_generators_active_profile_from_their_active_power()
            self.circuit.set_batteries_active_profile_from_their_active_power()

    def fix_loads_active_based_on_the_power(self, ask_before=True):
        """
        set the loads active based on the active power values
        :return:
        """

        if ask_before:
            ok = yes_no_question("This action sets the generation active profile based on the active power profile "
                                 "such that ig a generator active power is zero, the active value is false",
                                 "Set generation active profile")
        else:
            ok = True

        if ok:
            self.circuit.set_loads_active_profile_from_their_active_power()

    def get_preferred_engine(self) -> EngineType:
        """
        Get the currently selected engine
        :return: EngineType
        """
        val = self.ui.engineComboBox.currentText()
        return self.engine_dict[val]

    def about_box(self):
        """
        Display about box
        :return:
        """

        self.about_msg_window = AboutDialogueGuiGUI(self)
        self.about_msg_window.setVisible(True)

    @staticmethod
    def show_online_docs():
        """
        Open the online documentation in a web browser
        """
        webbrowser.open('https://veragrid.readthedocs.io/en/latest/', new=2)

    @staticmethod
    def report_a_bug():
        """
        Open the online github issues in a web browser
        """
        webbrowser.open('https://github.com/SanPen/VeraGrid/issues', new=2)

    def auto_rate_branches(self):
        """
        Rate the Branches that do not have rate
        """

        branches = self.circuit.get_branches()

        if len(branches) > 0:
            _, pf_results = self.session.power_flow

            if pf_results is not None:
                factor = self.ui.branch_rating_doubleSpinBox.value()

                for i, branch in enumerate(branches):

                    S = pf_results.Sf[i]

                    if branch.rate < 1e-3 or self.ui.rating_override_checkBox.isChecked():
                        r = np.round(abs(S) * factor, 1)
                        branch.rate = r if r > 0.0 else 1.0
                    else:
                        pass  # the rate is ok

            else:
                info_msg('Run a power flow simulation first.\nThe results are needed in this function.')

        else:
            warning_msg('There are no Branches!')

    def detect_transformers(self):
        """
        Detect which Branches are transformers
        """
        if len(self.circuit.lines) > 0:

            for elm in self.circuit.lines:

                v1 = elm.bus_from.Vnom
                v2 = elm.bus_to.Vnom

                if abs(v1 - v2) > 1.0:
                    self.circuit.convert_line_to_transformer(elm)
                else:

                    pass  # is a line

        else:
            warning_msg('There are no Branches!')

    def set_cancel_state(self) -> None:
        """
        Cancel what ever's going on that can be cancelled
        @return:
        """

        reply = QtWidgets.QMessageBox.question(self, 'Message',
                                               'Are you sure that you want to cancel the simulation?',
                                               QtWidgets.QMessageBox.StandardButton.Yes,
                                               QtWidgets.QMessageBox.StandardButton.No)

        if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:
            # send the cancel state to whatever it is being executed

            for drv in self.get_all_threads():
                if drv is not None:
                    if hasattr(drv, 'cancel'):
                        drv.cancel()
        else:
            pass

    def display_grid_analysis(self):
        """
        Display the grid analysis GUI
        """

        self.analysis_dialogue = GridAnalysisGUI(circuit=self.circuit)

        self.analysis_dialogue.resize(int(1.61 * 600.0), 600)
        self.analysis_dialogue.show()

    def change_circuit_base(self):
        """
        Update the circuit base values from the UI
        """

        Sbase_new = self.ui.sbase_doubleSpinBox.value()
        self.circuit.change_base(Sbase_new)

        self.circuit.fBase = self.ui.fbase_doubleSpinBox.value()

    def change_circuit_name(self):
        """

        :return:
        """
        self.circuit.name = self.ui.grid_name_line_edit.text().strip()

    def get_snapshot_circuit(self):
        """
        Get a snapshot compilation
        :return: SnapshotData instance
        """
        return compile_numerical_circuit_at(circuit=self.circuit)

    @property
    def numerical_circuit(self) -> NumericalCircuit:
        """
        get the snapshot NumericalCircuit
        :return: NumericalCircuit
        """
        return self.get_snapshot_circuit()

    @property
    def islands(self) -> List[NumericalCircuit]:
        """
        get the snapshot islands
        :return: List[NumericalCircuit]
        """
        numerical_circuit = compile_numerical_circuit_at(circuit=self.circuit)
        calculation_inputs = numerical_circuit.split_into_islands()
        return calculation_inputs

    def initialize_contingencies(self):
        """
        Launch the contingency planner to initialize the contingencies
        :return:
        """
        self.contingency_planner_dialogue = ContingencyPlannerGUI(parent=self, grid=self.circuit)
        self.contingency_planner_dialogue.exec()

        # gather results
        if self.contingency_planner_dialogue.generated_results:
            if len(self.contingency_planner_dialogue.contingency_groups):
                self.circuit.contingency_groups += self.contingency_planner_dialogue.contingency_groups
                self.circuit.contingencies += self.contingency_planner_dialogue.contingencies
            else:
                info_msg(text="No contingencies were generated :/", title="Contingency planner")

    def show_toast(self, message: str, duration: int = 2000):
        """
        Show generic toast
        :param message: Message to display
        :param duration: duration in ms
        """
        self.toast_manager.show_toast(message=message, duration=duration, toast_type="veragrid")

    def show_error_toast(self, message: str, duration: int = 2000):
        """
        Show error toast
        :param message: Message to display
        :param duration: duration in ms
        """
        self.toast_manager.show_error_toast(message=message, duration=duration)

    def show_warning_toast(self, message: str, duration: int = 2000):
        """
        Show warning toast
        :param message: Message to display
        :param duration: duration in ms
        """
        self.toast_manager.show_warning_toast(message=message, duration=duration)

    def show_info_toast(self, message: str, duration: int = 2000):
        """
        Show info toast
        :param message: Message to display
        :param duration: duration in ms
        """
        self.toast_manager.show_info_toast(message=message, duration=duration)

    def get_clustering_results(self) -> Union[ClusteringResults, None]:
        """
        Get the clustering results if available
        :return: ClusteringResults or None
        """
        if self.ui.actionUse_clustering.isChecked():
            _, clustering_results = self.session.clustering

            if clustering_results is not None:
                n = len(clustering_results.time_indices)

                if n != self.ui.cluster_number_spinBox.value():
                    error_msg("The number of clusters in the stored results is different from the specified :(\n"
                              "Run another clustering analysis.")

                    return None
                else:
                    # all ok
                    return clustering_results
            else:
                # no results ...
                self.show_warning_toast("There are no clustering results.")
                self.ui.actionUse_clustering.setChecked(False)
                return None

        else:
            # not marked ...
            return None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import TYPE_CHECKING

from VeraGridEngine.Devices.Injections.battery import Battery
from VeraGrid.Gui.Diagrams.MapWidget.Injections.map_injections_template_graphics import MapInjectionTemplateGraphicItem

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget


class MapBatteryGraphicItem(MapInjectionTemplateGraphicItem):

    def __init__(self, api_object: Battery,
                 editor: GridMapWidget,
                 lat: float,
                 lon: float,
                 size: float = 0.8,
                 draw_labels: bool = True):
        """

        :param api_object:
        :param editor:
        """
        MapInjectionTemplateGraphicItem.__init__(self,
                                                 api_object=api_object,
                                                 editor=editor,
                                                 lat=lat,
                                                 lon=lon,
                                                 size=size,
                                                 draw_labels=draw_labels)

    @property
    def api_object(self) -> Battery:
        """

        :return:
        """
        return self._api_object



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

from typing import List, Tuple, Dict
from collections import defaultdict
from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QGroupBox, QLabel,
    QComboBox, QCheckBox, QTableWidget, QTableWidgetItem, QMessageBox,
    QPushButton, QHeaderView, QStyledItemDelegate, QSpinBox, QAbstractItemView
)
from PySide6.QtCore import QModelIndex, QPersistentModelIndex, Qt
import VeraGridEngine.Devices as dev
from VeraGridEngine.enumerations import VoltageLevelTypes


class SpinBoxDelegate(QStyledItemDelegate):
    """
    Delegate for integer-only column (Calle).
    """

    def __init__(self, parent=None, minimum=1, maximum=9999):
        super().__init__(parent)
        self.minimum = minimum
        self.maximum = maximum

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        spinbox = QSpinBox(parent)
        spinbox.setMinimum(self.minimum)
        spinbox.setMaximum(self.maximum)
        return spinbox

    def setEditorData(self, editor: QSpinBox, index: QModelIndex | QPersistentModelIndex):
        """

        :param editor:
        :param index:
        """
        value = index.model().data(index, 0)
        if value and value.isdigit():
            editor.setValue(int(value))
        else:
            editor.setValue(self.minimum)

    def setModelData(self, editor: QSpinBox, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        model.setData(index, str(editor.value()))


class ComboBoxDelegate(QStyledItemDelegate):
    """Delegate for restricted values (Posición, Barra)."""

    def __init__(self, items: List[str], parent: QTableWidget = None) -> None:
        super().__init__(parent)
        self.items = items

    def createEditor(self, parent, option, index):
        """

        :param parent:
        :param option:
        :param index:
        :return:
        """
        combo = QComboBox(parent)
        combo.addItems(self.items)
        return combo

    def setEditorData(self, editor: QComboBox, index: QModelIndex | QPersistentModelIndex):
        """

        :param editor:
        :param index:
        """
        value = index.model().data(index, 0)
        if value in self.items:
            editor.setCurrentText(value)
        else:
            editor.setCurrentIndex(0)

    def setModelData(self, editor: QComboBox, model, index):
        """

        :param editor:
        :param model:
        :param index:
        """
        model.setData(index, editor.currentText())


class TableEntry:
    """
    Table entry
    """

    def __init__(self, device: str, bay: str, main_bar: str):
        self.device = device
        self.bay = bay
        self.main_bar = main_bar


def get_number_of_bars(vl: VoltageLevelTypes, n_bay: int):
    """
    Get the number of bars of the configuration
    :param vl: VoltageLevelTypes
    :param n_bay: number of bays (used for the ring config)
    :return: number of main bars
    """
    if vl == VoltageLevelTypes.SingleBar:
        return 1
    if vl == VoltageLevelTypes.SingleBarWithBypass:
        return 1
    if vl == VoltageLevelTypes.SingleBarWithSplitter:
        return 1
    if vl == VoltageLevelTypes.DoubleBar:
        return 2
    if vl == VoltageLevelTypes.DoubleBarWithBypass:
        return 2
    if vl == VoltageLevelTypes.DoubleBarWithTransference:
        return 2
    if vl == VoltageLevelTypes.DoubleBarDuplex:
        return 2
    if vl == VoltageLevelTypes.Ring:
        return n_bay
    if vl == VoltageLevelTypes.BreakerAndAHalf:
        return 2

    return 0


def is_main_bar_column_editable(vl: VoltageLevelTypes) -> bool:
    """
    Determine if the Main Bar column should be editable for the given voltage level type.
    Only Breaker-and-a-Half scheme requires user to specify which bus each position connects to.

    :param vl: VoltageLevelTypes
    :return: True if Main Bar column should be editable
    """
    return vl == VoltageLevelTypes.BreakerAndAHalf


class ReorderableTableWidget(QTableWidget):
    """
    A QTableWidget that supports row reordering via up/down buttons.
    """

    def __init__(self, rows: int, columns: int, parent=None):
        super().__init__(rows, columns, parent)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self._auto_update_bay_numbers = True  # Can be disabled for schemes like Breaker-and-a-Half

    def set_auto_update_bay_numbers(self, enabled: bool) -> None:
        """Enable/disable automatic bay number updates on reorder."""
        self._auto_update_bay_numbers = enabled

    def move_row_up(self) -> None:
        """Move the selected row up by one position."""
        current_row = self.currentRow()
        if current_row > 0:
            self._swap_rows(current_row, current_row - 1)
            self.selectRow(current_row - 1)
            if self._auto_update_bay_numbers:
                self._update_bay_numbers()

    def move_row_down(self) -> None:
        """Move the selected row down by one position."""
        current_row = self.currentRow()
        # if current_row < self.rowCount() - 1 and current_row >= 0:
        if self.rowCount() - 1 > current_row >= 0:
            self._swap_rows(current_row, current_row + 1)
            self.selectRow(current_row + 1)
            if self._auto_update_bay_numbers:
                self._update_bay_numbers()

    def _swap_rows(self, row1: int, row2: int) -> None:
        """Swap the contents of two rows."""
        if self._auto_update_bay_numbers:
            # Only swap device name (col 0) and assigned bus (col 2)
            # Bay number (col 1) stays in place and will be updated by _update_bay_numbers
            cols_to_swap = [0, 2]
        else:
            # Swap all columns (for schemes where user controls bay numbers)
            cols_to_swap = list(range(self.columnCount()))

        for col in cols_to_swap:
            item1 = self.takeItem(row1, col)
            item2 = self.takeItem(row2, col)
            if item1 and item2:
                self.setItem(row1, col, item2)
                self.setItem(row2, col, item1)
            elif item1:
                self.setItem(row2, col, item1)
            elif item2:
                self.setItem(row1, col, item2)

    def _update_bay_numbers(self) -> None:
        """Update bay numbers to match row positions."""
        bay_num = 1
        for row in range(self.rowCount()):
            bay_item = self.item(row, 1)
            if bay_item:
                # Skip visual indicator rows (like "Double bar connection" with "-")
                if bay_item.text() == "-":
                    continue
                bay_item.setText(f"Bay{bay_num}")
                bay_num += 1


class VoltageLevelConversionWizard(QDialog):
    """
    Voltage level conversion wizard for converting a bus-branch model to a node-breaker model.
    Supports Single Bus, Double Bus, Breaker-and-a-Half, and Ring configurations (as desired by REE)
    """

    def __init__(self, bus: dev.Bus, grid: dev.MultiCircuit):
        """
        Constructor
        :param bus: Bus to modify
        :param grid: MultiCircuit where the bus belongs
        """
        super().__init__()
        self.setWindowTitle("Convert Bus to Voltage Level")
        self.setMinimumWidth(600)
        self.setMinimumHeight(500)

        self.bus = bus
        self.grid = grid

        self.bus_branches, self.bus_injections = self.grid.get_bus_devices(self.bus)
        self.all_dev = self.bus_branches + self.bus_injections

        # Main layout
        main_layout = QVBoxLayout(self)

        # --- Combobox for type of scheme ---
        scheme_layout = QHBoxLayout()
        scheme_layout.addWidget(QLabel("Scheme type:"))

        # Only show the 4 supported schemes in the dropdown (as desired by REE)
        # Other schemes are kept in the backend but not exposed in the UI
        self.vl_list = [
            VoltageLevelTypes.SingleBar,
            VoltageLevelTypes.DoubleBar,
            VoltageLevelTypes.BreakerAndAHalf,
            VoltageLevelTypes.Ring,
        ]

        self.vl_dict = {vl.value: vl for vl in self.vl_list}

        self.combo = QComboBox()
        self.combo.addItems([vl.value for vl in self.vl_list])
        scheme_layout.addWidget(self.combo)
        scheme_layout.addStretch()
        main_layout.addLayout(scheme_layout)
        self.combo.currentIndexChanged.connect(self.reset_all)

        # Checkboxes in a group box
        options_group = QGroupBox("Options")
        options_layout = QVBoxLayout(options_group)

        # First row of checkboxes
        row1_layout = QHBoxLayout()
        self.add_breakers_checkbox = QCheckBox("Use breakers")
        self.add_breakers_checkbox.setChecked(True)
        self.add_breakers_checkbox.setToolTip("If disabled, connections are made using branches instead of breakers")
        row1_layout.addWidget(self.add_breakers_checkbox)

        self.keep_original_ratio_checkbox = QCheckBox("Keep original rates")
        self.keep_original_ratio_checkbox.setChecked(False)
        self.keep_original_ratio_checkbox.setToolTip("If enabled, breakers inherit the rates "
                                                     "of the original branch. Otherwise, 9999 is assigned")
        row1_layout.addWidget(self.keep_original_ratio_checkbox)
        row1_layout.addStretch()
        options_layout.addLayout(row1_layout)

        # Second row of checkboxes
        row2_layout = QHBoxLayout()
        self.enable_transfer_bus_checkbox = QCheckBox("Enable transfer bus (JBPT)")
        self.enable_transfer_bus_checkbox.setChecked(False)
        self.enable_transfer_bus_checkbox.setToolTip("If enabled, a transfer bus (JBPT) is added with additional logic")
        row2_layout.addWidget(self.enable_transfer_bus_checkbox)

        self.reducible_branches_checkbox = QCheckBox("Reducible branches")
        self.reducible_branches_checkbox.setChecked(False)
        self.reducible_branches_checkbox.setToolTip("If enabled, new branches are marked as reducible")
        row2_layout.addWidget(self.reducible_branches_checkbox)
        row2_layout.addStretch()
        options_layout.addLayout(row2_layout)

        # Third row of checkboxes
        row3_layout = QHBoxLayout()
        self.bar_by_segments_checkbox = QCheckBox("Bars with impedance")
        self.bar_by_segments_checkbox.setChecked(False)
        self.bar_by_segments_checkbox.setToolTip("Split the bar into segments with impedances")
        row3_layout.addWidget(self.bar_by_segments_checkbox)
        row3_layout.addStretch()
        options_layout.addLayout(row3_layout)

        main_layout.addWidget(options_group)

        # --- Table for positions ---
        table_label = QLabel("Positions (use arrows to reorder):")
        main_layout.addWidget(table_label)

        # Table with reorder buttons layout
        table_layout = QHBoxLayout()

        self.bays_list = list()
        self.main_bars_list = list()
        self.table = ReorderableTableWidget(len(self.all_dev), 3)
        self.table.setHorizontalHeaderLabels(["Device", "Bay", "Assigned Bus"])
        self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        table_layout.addWidget(self.table)

        # Reorder buttons (up/down arrows)
        reorder_buttons_layout = QVBoxLayout()
        reorder_buttons_layout.addStretch()

        self.move_up_button = QPushButton("\u25B2")  # Up arrow
        self.move_up_button.setFixedWidth(30)
        self.move_up_button.setToolTip("Move selected row up")
        self.move_up_button.clicked.connect(self.table.move_row_up)
        reorder_buttons_layout.addWidget(self.move_up_button)

        self.move_down_button = QPushButton("\u25BC")  # Down arrow
        self.move_down_button.setFixedWidth(30)
        self.move_down_button.setToolTip("Move selected row down")
        self.move_down_button.clicked.connect(self.table.move_row_down)
        reorder_buttons_layout.addWidget(self.move_down_button)

        reorder_buttons_layout.addStretch()
        table_layout.addLayout(reorder_buttons_layout)

        main_layout.addLayout(table_layout)

        # --- Add/Remove position buttons ---
        position_buttons_layout = QHBoxLayout()
        self.add_spare_button = QPushButton("+ Add spare position")
        self.add_spare_button.setToolTip("Add an empty position (especially important for Breaker-and-a-Half scheme)")
        self.add_spare_button.clicked.connect(self.add_spare_position)
        position_buttons_layout.addWidget(self.add_spare_button)

        self.remove_row_button = QPushButton("- Remove selected")
        self.remove_row_button.setToolTip("Remove the selected row")
        self.remove_row_button.clicked.connect(self.remove_selected_row)
        position_buttons_layout.addWidget(self.remove_row_button)

        position_buttons_layout.addStretch()
        main_layout.addLayout(position_buttons_layout)

        # --- Bottom buttons layout ---
        bottom_layout = QHBoxLayout()
        bottom_layout.addStretch()

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.close)
        bottom_layout.addWidget(self.cancel_button)

        self.do_button = QPushButton("Do it")
        self.do_button.clicked.connect(self.do_it)
        bottom_layout.addWidget(self.do_button)

        main_layout.addLayout(bottom_layout)

        self.closed_ok: bool = False

        self.reset_all()

    def reset_all(self) -> None:
        """
        Reset the table and configuration based on the selected scheme type.
        """
        vl_type: VoltageLevelTypes = self.get_vl_type()
        n_positions = len(self.bus_branches) + len(self.bus_injections)

        # For Breaker-and-a-Half, we need at least enough bays for all positions
        # (positions can share bays, max 2 per bay)
        if vl_type == VoltageLevelTypes.BreakerAndAHalf:
            n_bay = (n_positions + 1) // 2  # ceil division
        else:
            n_bay = n_positions

        n_bars = get_number_of_bars(vl=vl_type, n_bay=n_positions)

        # Allowed lists
        self.bays_list = [f"Bay{i + 1}" for i in range(max(n_bay * 2, n_positions + 5))]  # Extra bays for flexibility
        self.main_bars_list = ["JBP1", "JBP2"] if n_bars >= 2 else ["JBP1"]

        # Determine if main bar column should be editable
        main_bar_editable = is_main_bar_column_editable(vl_type)

        # For Breaker-and-a-Half, disable auto bay number update (user controls bay grouping)
        # For other schemes, enable auto update so reordering updates bay numbers
        self.table.set_auto_update_bay_numbers(vl_type != VoltageLevelTypes.BreakerAndAHalf)

        # Clear and rebuild table
        self.table.setRowCount(n_positions)

        # Collect device names
        dev_names: List[str] = list()

        # Fill data. For Breaker-and-a-Half, alternate bus assignments in pairs
        data: List[TableEntry] = list()
        for i, elm in enumerate(self.bus_branches + self.bus_injections):
            if vl_type == VoltageLevelTypes.BreakerAndAHalf:
                # Pair positions in bays, alternating bus assignments
                bay_num = (i // 2) + 1
                # First position in bay gets JBP1, second gets JBP2
                assigned_bus = "JBP1" if (i % 2 == 0) else "JBP2"
            else:
                bay_num = i + 1
                assigned_bus = "JBP1"

            name = f"{elm.name} ({elm.device_type.value})"
            data.append(TableEntry(name, f"Bay{bay_num}", assigned_bus))
            dev_names.append(name)

        dev_names.append("(Spare)")  # Always allow spare as an option

        # Apply delegates
        self.table.setItemDelegateForColumn(0, ComboBoxDelegate(dev_names, self.table))
        self.table.setItemDelegateForColumn(1, ComboBoxDelegate(self.bays_list, self.table))
        self.table.setItemDelegateForColumn(2, ComboBoxDelegate(self.main_bars_list, self.table))

        for row, entry in enumerate(data):
            device_item = QTableWidgetItem(entry.device)
            bay_item = QTableWidgetItem(entry.bay)
            main_bar_item = QTableWidgetItem(entry.main_bar)

            # Make main bar column read-only if not editable for this scheme
            if not main_bar_editable:
                main_bar_item.setFlags(main_bar_item.flags() & ~Qt.ItemFlag.ItemIsEditable)
                main_bar_item.setBackground(Qt.GlobalColor.lightGray)

            self.table.setItem(row, 0, device_item)
            self.table.setItem(row, 1, bay_item)
            self.table.setItem(row, 2, main_bar_item)

        # Add visual indicator row for Double bar scheme
        if vl_type == VoltageLevelTypes.DoubleBar:
            connection_row = self.table.rowCount()
            self.table.insertRow(connection_row)

            connection_device_item = QTableWidgetItem("Double bar connection")
            connection_bay_item = QTableWidgetItem("-")
            connection_bus_item = QTableWidgetItem("JBP1")

            # Make all cells non-editable and non-selectable for this visual indicator row
            non_interactive_flags = Qt.ItemFlag.ItemIsEnabled  # Only enabled, not editable or selectable
            connection_device_item.setFlags(non_interactive_flags)
            connection_bay_item.setFlags(non_interactive_flags)
            connection_bus_item.setFlags(non_interactive_flags)

            self.table.setItem(connection_row, 0, connection_device_item)
            self.table.setItem(connection_row, 1, connection_bay_item)
            self.table.setItem(connection_row, 2, connection_bus_item)

    def add_spare_position(self) -> None:
        """
        Add a spare position row to the table.
        This is especially useful for Breaker-and-a-Half scheme where bays may need to be completed.
        """
        vl_type = self.get_vl_type()
        main_bar_editable = is_main_bar_column_editable(vl_type)

        row = self.table.rowCount()
        self.table.insertRow(row)

        # Determine next bay number
        next_bay = row + 1

        # Set items
        device_item = QTableWidgetItem("(Spare)")
        bay_item = QTableWidgetItem(f"Bay{next_bay}")
        main_bar_item = QTableWidgetItem("JBP1")

        if not main_bar_editable:
            main_bar_item.setFlags(main_bar_item.flags() & ~Qt.ItemFlag.ItemIsEditable)
            main_bar_item.setBackground(Qt.GlobalColor.lightGray)

        self.table.setItem(row, 0, device_item)
        self.table.setItem(row, 1, bay_item)
        self.table.setItem(row, 2, main_bar_item)

    def remove_selected_row(self) -> None:
        """
        Remove the currently selected row from the table
        Only allows removal of spare positions (not actual devices)
        """
        current_row = self.table.currentRow()
        if current_row >= 0:
            device_item = self.table.item(current_row, 0)
            if device_item and device_item.text() == "(Spare)":
                self.table.removeRow(current_row)
            else:
                QMessageBox.warning(
                    self,
                    "Cannot Remove",
                    "Only spare positions can be removed. Actual devices cannot be removed from the list."
                )

    def get_vl_type(self) -> VoltageLevelTypes:
        """
        Get the currently selected voltage level type.
        :return: VoltageLevelTypes enum value
        """
        return self.vl_dict[self.combo.currentText()]

    def validate_configuration(self) -> Tuple[bool, str]:
        """
        Validate the current configuration based on the selected scheme type.
        :return: Tuple of (is_valid, error_message)
        """
        vl_type = self.get_vl_type()
        bay_assignments = self.get_bay_assignments()

        # Group by bay
        bays: Dict[int, List[Tuple[str, str]]] = defaultdict(list)
        for device, bay_num, bus in bay_assignments:
            bays[bay_num].append((device, bus))

        if vl_type == VoltageLevelTypes.BreakerAndAHalf:
            # Check Breaker-and-a-Half specific rules
            for bay_num, positions in bays.items():
                if len(positions) > 2:
                    return False, f"Bay {bay_num} has {len(positions)} positions. Maximum is 2 for Breaker-and-a-Half."

                if len(positions) == 2:
                    # Both positions must have different bus assignments
                    buses = [p[1] for p in positions]
                    if buses[0] == buses[1]:
                        return False, f"Bay {bay_num} has two positions with the same bus assignment ({buses[0]}). Each position must connect to a different bus."
        else:
            # For other schemes, check no duplicate bay numbers
            for bay_num, positions in bays.items():
                if len(positions) > 1:
                    return False, f"Bay {bay_num} has {len(positions)} positions. Only 1 position per bay is allowed for {vl_type.value}."

        return True, ""

    def get_bay_assignments(self) -> List[Tuple[str, int, str]]:
        """
        Get the bay assignments from the table.
        :return: List of (device_name, bay_number, assigned_bus) tuples
                 Note: device_name is the plain device name (without device type suffix)
        """
        assignments = []
        for row in range(self.table.rowCount()):
            device_item = self.table.item(row, 0)
            bay_item = self.table.item(row, 1)
            main_bar_item = self.table.item(row, 2)

            device_display = device_item.text() if device_item else ""

            # Skip visual indicator rows (like "Double bar connection")
            if device_display == "Double bar connection":
                continue

            # Extract plain device name from display format "DeviceName (DeviceType)"
            # Keep "(Spare)" as-is since it's a special marker
            if device_display == "(Spare)":
                device_name = device_display
            elif " (" in device_display and device_display.endswith(")"):
                # Extract name before the " (DeviceType)" suffix
                device_name = device_display.rsplit(" (", 1)[0]
            else:
                device_name = device_display

            bay_str = bay_item.text() if bay_item else "Bay1"
            main_bar = main_bar_item.text() if main_bar_item else "JBP1"

            # Extract bay number from string like "Bay1"
            try:
                bay_num = int(bay_str.replace("Bay", ""))
            except ValueError:
                bay_num = 1

            assignments.append((device_name, bay_num, main_bar))

        return assignments

    def get_device_to_bay_mapping(self) -> Dict[str, Tuple[int, str]]:
        """
        Get a mapping from device name to (bay_number, assigned_bus).
        :return: Dictionary mapping device name to (bay, bus) tuple
        """
        mapping = {}
        for device, bay, bus in self.get_bay_assignments():
            if device != "(Spare)":
                mapping[device] = (bay, bus)
        return mapping

    def do_it(self) -> None:
        """
        Validate configuration and perform the conversion if valid.
        """
        is_valid, error_msg = self.validate_configuration()

        if not is_valid:
            QMessageBox.warning(self, "Validation Error", error_msg)
            return

        self.closed_ok = True
        self.close()


# if __name__ == "__main__":
#     import sys
#     from PySide6.QtWidgets import QApplication
#     import VeraGridEngine as vg
#
#     # fname = "/home/santi/Documentos/Git/GitHub/VeraGrid_bkup/src/tests/data/grids/lynn5node.gridcal"
#     fname = "/Users/santi/Git/eRoots/VeraGrid/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal"
#     _grid = vg.open_file(fname)
#
#     app = QApplication(sys.argv)
#     window = VoltageLevelConversionWizard(grid=_grid, bus=_grid.buses[2])
#     window.resize(600, 500)
#     window.show()
#     sys.exit(app.exec())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

import os.path
import sys

from PySide6 import QtWidgets, QtGui
from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from VeraGrid.Gui.Main.MainWindow import QApplication
from PySide6.QtCore import QDirIterator, QResource
from PySide6.QtSvg import QSvgRenderer
from VeraGrid.Gui.update_gui_all import update_all_icons
from VeraGrid.Gui.Main.SubClasses.Scripting.scripting import ScriptingMain
import VeraGrid.ThirdParty.qdarktheme as qdarktheme
from VeraGrid.__version__ import __VeraGrid_VERSION__

__author__ = 'Santiago Peñate Vera'

"""
This class is the handler of the main gui of VeraGrid.
"""


########################################################################################################################
# Main Window
########################################################################################################################

class VeraGridMainGUI(ScriptingMain):
    """
    MainGUI
    """

    def __init__(self) -> None:
        """
        Main constructor
        """

        # create main window
        ScriptingMain.__init__(self, parent=None)
        self.setWindowTitle('VeraGrid ' + __VeraGrid_VERSION__)
        self.setAcceptDrops(True)

        self.ui.mainTabWidget.setCurrentIndex(0)
        self.ui.modelTabWidget.setCurrentIndex(0)
        self.ui.settingsTabWidget.setCurrentIndex(0)
        self.ui.resultsTabWidget.setCurrentIndex(0)

        ################################################################################################################
        # Set splitters
        ################################################################################################################

        # 1:4
        self.ui.dataStructuresSplitter.setStretchFactor(0, 3)
        self.ui.dataStructuresSplitter.setStretchFactor(1, 4)

        self.ui.simulationDataSplitter.setStretchFactor(1, 15)

        self.ui.results_splitter.setStretchFactor(0, 2)
        self.ui.results_splitter.setStretchFactor(1, 4)

        self.ui.diagram_selection_splitter.setStretchFactor(0, 10)
        self.ui.diagram_selection_splitter.setStretchFactor(1, 2)

        ################################################################################################################
        # Other actions
        ################################################################################################################

        self.ui.grid_colouring_frame.setVisible(True)

        self.ui.actionSync.setVisible(False)

        self.modify_ui_options_according_to_the_engine()

        # this is the contingency planner tab, invisible until done
        self.ui.modelTabWidget.setTabVisible(4, True)

        self.clear_results()

        self.load_all_config()

        self.add_complete_bus_branch_diagram()
        # self.add_map_diagram(ask=False)
        self.set_diagram_widget(self.diagram_widgets_list[0])
        self.update_available_results()

        self.ui.actionRun_Dynamic_RMS_Simulation.setVisible(True)
        self.ui.actionRun_Small_Signal_RMS_Simulation.setVisible(True)

        # global delete function
        self.ui.actionDelete_selected.triggered.connect(self.global_delete)

    def global_delete(self):
        """
        Function to dispatch what to do when [supr] is pressed
        :return:
        """
        if self.ui.mainTabWidget.currentIndex() == 0:  # Model
            if self.ui.modelTabWidget.currentIndex() == 0:  # Diagrams
                self.delete_selected_diagram_widgets()

            elif self.ui.modelTabWidget.currentIndex() == 1:  # Database
                self.delete_selected_db_table_objects()

            else:
                self.show_warning_toast("No effect, select diagrams or database")
        else:
            self.show_warning_toast("No effect, select diagrams or database")

    def save_all_config(self) -> None:
        """
        Save all configuration files needed
        """
        self.save_gui_config()
        self.save_server_config()

    def load_all_config(self) -> None:
        """
        Load all configuration files needed
        """
        self.load_gui_config()
        self.load_server_config()
        self.add_plugins()

        # apply the theme selected by the settings
        self.change_theme_mode()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """
        Close event
        :param event:
        :return:
        """
        if self.circuit.get_bus_number() > 0:
            quit_msg = "Are you sure that you want to exit VeraGrid?"
            reply = QtWidgets.QMessageBox.question(self, 'Close', quit_msg,
                                                   QtWidgets.QMessageBox.StandardButton.Yes,
                                                   QtWidgets.QMessageBox.StandardButton.No)

            if reply == QtWidgets.QMessageBox.StandardButton.Yes:
                # save config regardless
                self.save_all_config()
                self.stop_all_threads()
                event.accept()
            else:
                # save config regardless
                self.save_all_config()
                event.ignore()
        else:
            # no buses so exit
            # save config regardless
            self.save_all_config()
            self.stop_all_threads()
            event.accept()


def create_linux_desktop_entry(app_name: str, qrc_icon_path: str):
    """
    Create a .desktop entry for a PySide app using a resource icon (":/path/to/icon.svg").

    Parameters
    ----------
    app_name : str
        Name of the application (also used for StartupWMClass).
    qrc_icon_path : str
        Path to the icon inside the .qrc (e.g. ':/icons/app_icon.svg')
    """
    if not sys.platform.startswith("linux"):
        # print("[INFO] Not running on Linux, skipping .desktop creation.")
        return None

    # Extract icon from Qt resource to a real file
    icon = QIcon(qrc_icon_path)
    if icon.isNull():
        print(f"[WARNING] Could not find icon in resource: {qrc_icon_path}")
        return None

    # Temporary export path (user local cache)
    cache_dir = os.path.expanduser(f"~/.cache/{app_name}")
    os.makedirs(cache_dir, exist_ok=True)
    icon_path = os.path.join(cache_dir, f"{app_name}.png")

    # Save first available icon size
    pixmap = icon.pixmap(256, 256)
    pixmap.save(icon_path, "PNG")

    # Create .desktop entry
    desktop_dir = os.path.expanduser("~/.local/share/applications")
    os.makedirs(desktop_dir, exist_ok=True)
    desktop_file = os.path.join(desktop_dir, f"{app_name}.desktop")

    if not os.path.exists(desktop_file):
        exec_path = f"{sys.executable} {os.path.abspath(sys.argv[0])}"
        content = f"""[Desktop Entry]
Version={__VeraGrid_VERSION__}
Type=Application
Name={app_name}
Exec={exec_path}
Icon={icon_path}
Terminal=false
StartupWMClass={app_name}
Categories=Utility;
"""
        with open(desktop_file, "w") as f:
            f.write(content)
        os.chmod(desktop_file, 0o755)
        print(f"[OK] Created .desktop entry: {desktop_file}")
    else:
        pass
        # print(f"[INFO] .desktop entry already exists: {desktop_file}")

    return desktop_file


def check_all_svgs():
    """
    Iterate through all resources registered by icons_rc and check SVG validity.
    :returns: if any icon has errors
    """
    it = QDirIterator(":/Icons", QDirIterator.IteratorFlag.Subdirectories)
    bad_files = []

    while it.hasNext():
        path = it.next()
        if path.lower().endswith(".svg"):
            res = QResource(path)
            if not res.isValid():
                print(f"[MISSING] {path}")
                bad_files.append((path, "missing"))
                continue

            renderer = QSvgRenderer(path)
            if not renderer.isValid():
                print(f"[INVALID] {path}")
                bad_files.append((path, "invalid"))
            else:
                pass

    if len(bad_files) > 0:
        print("\nSVG compatibility summary:")
        for path, status in bad_files:
            print(f"  {status.upper():<8} {path}")

        update_all_icons()
    else:
        print("SVG compatibility summary: all ok")


def runVeraGrid() -> None:
    """
    Main function to run the GUI
    :return:
    """

    # if hasattr(qdarktheme, 'enable_hi_dpi'):
    qdarktheme.enable_hi_dpi()

    app = QApplication(sys.argv)

    # MacOS: display icons in menus
    app.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus, False)

    icon_name = ':/Program icon/icons/VeraGrid_icon.png'
    icon = QtGui.QIcon(icon_name)

    # will check os internally
    create_linux_desktop_entry("veragrid", qrc_icon_path=icon_name)

    # MacOS: Fix to show the icon on the task bar
    app.setWindowIcon(icon)

    window_ = VeraGridMainGUI()
    window_.setWindowIcon(icon)  # also apply directly

    # process the argument if provided
    if len(sys.argv) > 1:
        f_name = sys.argv[1]
        if os.path.exists(f_name):
            window_.open_file_now(filenames=[f_name])

    # launch
    h_ = 780
    window_.resize(int(1.7 * h_), h_)  # almost the golden ratio :)
    window_.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    runVeraGrid()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
"""
A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
import os
import shutil
import platform
import subprocess
from setuptools import setup, find_packages
from setuptools.command.install import install

from VeraGrid.__version__ import __VeraGrid_VERSION__

here = os.path.abspath(os.path.dirname(__file__))

long_description = '''# VeraGrid

This software aims to be a complete platform for power systems research and simulation.

[Watch the video https](https://youtu.be/SY66WgLGo54)

[Check out the documentation](https://veragrid.readthedocs.io)


## Installation

pip install veragrid

For more options (including a standalone setup one), follow the
[installation instructions]( https://veragrid.readthedocs.io/en/latest/getting_started/install.html)
from the project's [documentation](https://veragrid.readthedocs.io)
'''

description = 'VeraGrid is a Power Systems simulation program intended for professional use and research'

base_path = os.path.join('VeraGrid')

pkgs_to_exclude = ['docs', 'research', 'tests', 'tutorials', 'VeraGridEngine']

packages = find_packages(exclude=pkgs_to_exclude)

# ... so we have to do the filtering ourselves
packages2 = list()
for package in packages:
    elms = package.split('.')
    excluded = False
    for exclude in pkgs_to_exclude:
        if exclude in elms:
            excluded = True

    if not excluded:
        packages2.append(package)

package_data = {'VeraGrid': ['*.md',
                            '*.rst',
                            'LICENSE.txt',
                            'setup.py',
                            'data/cables.csv',
                            'data/transformers.csv',
                            'data/wires.csv',
                            'data/sequence_lines.csv',
                            'data/VeraGrid.svg',
                            'data/VeraGrid.ico'],
                }

dependencies = ['setuptools>=41.0.1',
                'wheel>=0.37.2',
                "PySide6>=6.8.0",  # 5.14 breaks the UI generation for development, 6.7.0 breaks all
                "pytest>=7.2",
                "websockets",
                "opencv-python>=4.10.0.84",
                "packaging",
                "VeraGridEngine==" + __VeraGrid_VERSION__,  # the VeraGridEngine version must be exactly the same
                ]

extras_require = {
    'gch5 files': ["tables"]  # this is for h5 compatibility
}


class CustomInstall(install):
    def run(self):
        install.run(self)

        try:
            if platform.system() == "Linux":
                self.create_linux_menu_entry()
            elif platform.system() == "Windows":
                self.create_windows_shortcut()
            elif platform.system() == "macOS":
                self.create_macos_menu_entry()
            else:
                print("Unsupported OS: Installation might be incomplete.")
        except Exception as e:
            print("error creating the menu entry...", e)

    @staticmethod
    def create_linux_menu_entry():
        """

        :return:
        """
        icon_src = "data/VeraGrid.svg"

        # copy the icon
        dest_dir = f"~/.local/share/icons/hicolor/scalable/apps"
        os.makedirs(dest_dir, exist_ok=True)
        shutil.copy(icon_src, os.path.join(dest_dir, "VeraGrid.svg"))

        # Update the icon cache
        os.system("gtk-update-icon-cache /usr/share/icons/hicolor")

        desktop_entry = """
                [Desktop Entry]
                Name=VeraGrid
                Comment=Power systems done right
                Exec=veragrid
                Icon=~/.local/share/icons/hicolor/scalable/apps/VeraGrid.svg
                Terminal=false
                Type=Application
                Categories=Utility;Development;
                """
        path = os.path.expanduser("~/.local/share/applications/veragrid.desktop")
        with open(path, "w") as f:
            f.write(desktop_entry)
        os.chmod(path, 0o755)

    @staticmethod
    def create_windows_shortcut():
        """

        :return:
        """
        desktop = os.path.join(os.environ["USERPROFILE"], "Desktop")
        shortcut_path = os.path.join(desktop, "VeraGrid.lnk")

        # Use powershell to create a shortcut
        powershell_command = f"""
        $WScriptShell = New-Object -ComObject WScript.Shell;
        $Shortcut = $WScriptShell.CreateShortcut("{shortcut_path}");
        $Shortcut.TargetPath = "veragrid";
        $Shortcut.Arguments = "";
        $Shortcut.IconLocation = "path/to/your/icon.ico";
        $Shortcut.Save();
        """
        subprocess.run(["powershell", "-Command", powershell_command], check=True)
        print(f"Windows shortcut created at: {shortcut_path}")

    @staticmethod
    def create_macos_menu_entry():
        """

        :return:
        """
        svg_icon_src = "data/VeraGrid.svg"
        dest_dir = "/Applications/VeraGrid/icons/"
        os.makedirs(dest_dir, exist_ok=True)
        shutil.copy(svg_icon_src, os.path.join(dest_dir, "VeraGrid.svg"))
        print("macOS icon installed.")

        app_dir = "/Applications/VeraGrid.app"
        os.makedirs(app_dir, exist_ok=True)
        plist_content = """
        <?xml version="1.0" encoding="UTF-8"?>
        <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
        <plist version="1.0">
        <dict>
            <key>CFBundleName</key>
            <string>VeraGrid</string>
            <key>CFBundleIconFile</key>
            <string>VeraGrid</string>
            <key>CFBundleExecutable</key>
            <string>run_veragrid</string>
        </dict>
        </plist>
        """
        plist_path = os.path.join(app_dir, "Info.plist")
        with open(plist_path, "w") as f:
            f.write(plist_content)

        script_path = os.path.join(app_dir, "run_veragrid")
        with open(script_path, "w") as f:
            f.write("#!/bin/bash\nveragrid\n")
        os.chmod(script_path, 0o755)
        print(f"macOS application bundle created at {app_dir}.")


# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='VeraGrid',  # Required
    version=__VeraGrid_VERSION__,  # Required
    license='MPL2',
    description=description,  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://github.com/SanPen/VeraGrid',  # Optional
    author='Santiago Peñate Vera et. Al.',  # Optional
    author_email='spenate@eroots.tech',  # Optional
    classifiers=[
        'Programming Language :: Python :: 3.8',
    ],
    keywords='power systems planning',  # Optional
    packages=packages2,  # Required
    include_package_data=True,
    python_requires='>=3.8',
    install_requires=dependencies,
    extras_require=extras_require,
    package_data=package_data,
    entry_points={
        'console_scripts': [
            'veragrid = VeraGrid.ExecuteVeraGrid:runVeraGrid',
            'VeraGrid = VeraGrid.ExecuteVeraGrid:runVeraGrid',
        ],
    },
    cmdclass={'install': CustomInstall},
)



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""A module containing multiple filters used by template engine."""
from __future__ import annotations

import platform

from VeraGrid.ThirdParty.qdarktheme import __version__
from VeraGrid.ThirdParty.qdarktheme._color import Color
from VeraGrid.ThirdParty.qdarktheme._icon.svg import Svg
from VeraGrid.ThirdParty.qdarktheme._util import analyze_version_str, get_cash_root_path, get_logger
from VeraGrid.ThirdParty.qdarktheme.qtpy import __version__ as qt_version
from VeraGrid.ThirdParty.qdarktheme.qtpy.qt_compat import QT_API

_logger = get_logger(__name__)

if qt_version is None:
    _logger.warning("Failed to detect Qt version. Load Qt version as the latest version.")
    _QT_VERSION = "10.0.0"  # Fairly future version for always setting latest version.
else:
    _QT_VERSION = qt_version

if QT_API is None:
    _QT_API = "PySide6"
    _logger.warning(f"Failed to detect Qt binding. Load Qt API as '{_QT_API}'.")
else:
    _QT_API = QT_API

if None in (qt_version, QT_API):
    _logger.warning(
        "Maybe you need to install qt-binding. "
        "Available Qt-binding packages: PySide6, PyQt6, PyQt5, PySide2."
    )


def _transform(color: Color, color_state: dict[str, float]) -> Color:
    if color_state.get("transparent"):
        color = color.transparent(color_state["transparent"])
    if color_state.get("darken"):
        color = color.darken(color_state["darken"])
    if color_state.get("lighten"):
        return color.lighten(color_state["lighten"])
    return color


def color(color_info: str | dict[str, str | dict], state: str | None = None) -> Color:
    """Filter for template engine. This filter convert color info data to color object."""
    if isinstance(color_info, str):
        return Color.from_hex(color_info)

    base_color_format: str = color_info["base"]  # type: ignore
    color = Color.from_hex(base_color_format)

    if state is None:
        return color

    transforms = color_info[state]
    return Color.from_hex(transforms) if isinstance(transforms, str) else _transform(color, transforms)


def palette_format(color: Color) -> str:
    """Filter for template engine. This filter convert color object to ARGB hex format.

    QPalette parser for hex only support ARGB hex format. color.Color class use RGB hex format.
    So we need to convert Color object to ARGB hex format.
    """
    return f"#{color.to_hex_argb()}"


def url(color: Color, id: str, rotate: int = 0) -> str:
    """Filter for template engine. This filter create url for svg and output svg file."""
    svg_path = get_cash_root_path(__version__) / f"{id}_{color._to_hex()}_{rotate}.svg"
    url = f"url({svg_path.as_posix()})"
    if svg_path.exists():
        return url
    svg = Svg(id).colored(color).rotate(rotate)
    svg_path.write_text(str(svg))
    return url


def env(
    text, value: str, version: str | None = None, qt: str | None = None, os: str | None = None
) -> str:
    """Filter for template engine. This filter output empty string when unexpected environment."""
    if version and not analyze_version_str(_QT_VERSION, version):
        return ""
    if qt and qt.lower() != _QT_API.lower():
        return ""
    if os and platform.system().lower() not in os.lower():
        return ""
    return value.replace("${}", str(text))


def corner(corner_shape: str, size: str) -> str:
    """Filter for template engine. This filter manage corner shape."""
    return size if corner_shape == "rounded" else "0"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
from typing import Dict, List, Union
from PySide6 import QtCore, QtWidgets, QtGui
from enum import EnumMeta
from VeraGrid.Gui.gui_functions import (TextDelegate, ColorPickerDelegate)
from VeraGridEngine.Devices import Bus, ContingencyGroup
from VeraGridEngine.Devices.Parents.editable_device import GCProp
from VeraGridEngine.enumerations import DeviceType
from VeraGrid.Gui.gui_functions import (ComboDelegate, FloatDelegate, ComplexDelegate)
from VeraGridEngine.Devices.types import ALL_DEV_TYPES


class RosetaObjectsModel(QtCore.QAbstractTableModel):
    """
    Class to populate a Qt table view with the properties of objects
    """

    def __init__(self,
                 objects,
                 editable_headers,
                 parent=None,
                 editable=False,
                 non_editable_attributes: Union[None, List[str]] = None,
                 transposed=False,
                 check_unique: Union[None, List[str]] = None,
                 dictionary_of_lists: Union[None, Dict[str, List[ALL_DEV_TYPES]]] = None):
        """

        :param objects: list of objects associated to the editor
        :param editable_headers: Dictionary with the properties and the units and type {attribute: ('unit', type)}
        :param parent: Parent object: the QTableView object
        :param editable: Is the table editable?
        :param non_editable_attributes: List of attributes that are not enabled for editing
        :param transposed: Display the table transposed?
        :param dictionary_of_lists: dictionary of lists for the Delegates
        """
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.parent = parent

        self.attributes = list(editable_headers.keys())

        self.attribute_types = [editable_headers[attr].class_type for attr in self.attributes]

        self.units = [editable_headers[attr].get_unit() for attr in self.attributes]

        self.tips = [editable_headers[attr].description for attr in self.attributes]

        self.objects = objects

        self.editable = editable

        self.non_editable_attributes = non_editable_attributes if non_editable_attributes is not None else list()

        self.check_unique = check_unique if check_unique is not None else list()

        self.r = len(self.objects)

        self.c = len(self.attributes)

        self.formatter = lambda x: "%.2f" % x

        self.transposed = transposed

        self.dictionary_of_lists = dictionary_of_lists if dictionary_of_lists is not None else dict()

        self.set_delegates()

    def set_delegates(self):
        """
        Set the cell editor types depending on the attribute_types array
        :return:
        """

        if self.transposed:
            F = self.parent.setItemDelegateForRow
        else:
            F = self.parent.setItemDelegateForColumn

        for i in range(self.c):
            tpe = self.attribute_types[i]

            if tpe is bool:
                delegate = ComboDelegate(self.parent, [True, False], ['True', 'False'])
                F(i, delegate)

            elif tpe is float:
                delegate = FloatDelegate(self.parent)
                F(i, delegate)

            elif tpe is complex:
                delegate = ComplexDelegate(self.parent)
                F(i, delegate)

            elif tpe is None:
                F(i, None)
                if len(self.non_editable_attributes) == 0:
                    self.non_editable_attributes.append(self.attributes[i])

            elif isinstance(tpe, EnumMeta):
                objects = list(tpe)
                values = [x.value for x in objects]
                delegate = ComboDelegate(self.parent, objects, values)
                F(i, delegate)

            # elif tpe in []:
            #
            #     objects = self.dictionary_of_lists[tpe.value]
            #     values = [x.name for x in objects]
            #     delegate = ComboDelegate(self.parent, objects, values)
            #     F(i, delegate)

            else:
                F(i, None)

    def update(self):
        """
        update table
        """
        row = self.rowCount()
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        # whatever code
        self.endInsertRows()

    def flags(self, index):
        """
        Get the display mode
        :param index:
        :return:
        """
        if self.transposed:
            attr_idx = index.row()
        else:
            attr_idx = index.column()

        if self.editable and self.attributes[attr_idx] not in self.non_editable_attributes:
            return QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def rowCount(self, parent=None):
        """
        Get number of rows
        :param parent:
        :return:
        """
        if self.transposed:
            return self.c
        else:
            return self.r

    def columnCount(self, parent=None):
        """
        Get number of columns
        :param parent:
        :return:
        """
        if self.transposed:
            return self.r
        else:
            return self.c

    def data_raw(self, r: int, c: int):
        """
        Get the data to display
        :param r: row index
        :param c: col index
        :return:
        """

        if self.transposed:
            obj_idx = c
            attr_idx = r
        else:
            obj_idx = r
            attr_idx = c

        attr = self.attributes[attr_idx]

        return getattr(self.objects[obj_idx], attr)

    def data_with_type(self, index):
        """
        Get the data to display
        :param index:
        :return:
        """

        if self.transposed:
            obj_idx = index.column()
            attr_idx = index.row()
        else:
            obj_idx = index.row()
            attr_idx = index.column()

        attr = self.attributes[attr_idx]

        return getattr(self.objects[obj_idx], attr)

    def data(self, index, role=None):
        """
        Get the data to display
        :param index:
        :param role:
        :return:
        """
        if index.isValid() and role == QtCore.Qt.ItemDataRole.DisplayRole:
            return str(self.data_with_type(index))

        return None

    def setData(self, index: QtCore.QModelIndex, value, role=None):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        :return:
        """

        if self.transposed:
            obj_idx = index.column()
            attr_idx = index.row()
        else:
            obj_idx = index.row()
            attr_idx = index.column()

        # check taken values
        if self.attributes[attr_idx] in self.check_unique:
            taken = self.attr_taken(self.attributes[attr_idx], value)
        else:
            taken = False

        if not taken:
            if self.attributes[attr_idx] not in self.non_editable_attributes:
                setattr(self.objects[obj_idx], self.attributes[attr_idx], value)
            else:
                pass  # the column cannot be edited

        return True

    def attr_taken(self, attr, val):
        """
        Checks if the attribute value is taken
        :param attr:
        :param val:
        :return:
        """
        for obj in self.objects:
            if val == getattr(obj, attr):
                return True
        return False

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """
        Get the headers to display
        :param section:
        :param orientation:
        :param role:
        :return:
        """
        if role == QtCore.Qt.ItemDataRole.DisplayRole:

            if self.transposed:
                # for the properties in the schematic view
                if orientation == QtCore.Qt.Orientation.Horizontal:
                    return 'Value'
                elif orientation == QtCore.Qt.Orientation.Vertical:
                    if self.units[section] != '':
                        return self.attributes[section]  # + ' [' + self.units[p_int] + ']'
                    else:
                        return self.attributes[section]
            else:
                # Normal
                if orientation == QtCore.Qt.Orientation.Horizontal:
                    if self.units[section] != '':
                        return self.attributes[section]  # + ' [' + self.units[p_int] + ']'
                    else:
                        return self.attributes[section]
                elif orientation == QtCore.Qt.Orientation.Vertical:
                    return str(section)  # + ':' + str(self.objects[p_int])

        # add a tooltip
        if role == QtCore.Qt.ItemDataRole.ToolTipRole:
            if section < self.c:
                if self.units[section] != "":
                    unit = '\nUnits: ' + self.units[section]
                else:
                    unit = ''
                return self.attributes[section] + unit + ' \n' + self.tips[section]
            else:
                # somehow the index is out of range
                return ""

        return None

    def copy_to_column(self, index):
        """
        Copy the value pointed by the index to all the other cells in the column
        :param index: QModelIndex instance
        :return:
        """
        value = self.data_with_type(index=index)
        col = index.column()

        for row in range(self.rowCount()):

            if self.transposed:
                obj_idx = col
                attr_idx = row
            else:
                obj_idx = row
                attr_idx = col

            if self.attributes[attr_idx] not in self.non_editable_attributes:
                setattr(self.objects[obj_idx], self.attributes[attr_idx], value)
            else:
                pass  # the column cannot be edited

    def get_data(self):
        """

        :return:
        """
        nrows = self.rowCount()
        ncols = self.columnCount()
        data = np.empty((nrows, ncols), dtype=object)

        for j in range(ncols):
            for i in range(nrows):
                data[i, j] = self.data_raw(r=i, c=j)

        columns = [self.headerData(i, orientation=QtCore.Qt.Orientation.Horizontal,
                                   role=QtCore.Qt.ItemDataRole.DisplayRole) for i in range(ncols)]

        index = [self.headerData(i, orientation=QtCore.Qt.Orientation.Vertical,
                                 role=QtCore.Qt.ItemDataRole.DisplayRole) for i in range(nrows)]

        return index, columns, data

    def copy_to_clipboard(self):
        """

        :return:
        """
        if self.columnCount() > 0:

            index, columns, data = self.get_data()

            data = data.astype(str)

            # header first
            txt = '\t' + '\t'.join(columns) + '\n'

            # data
            for t, index_value in enumerate(index):
                txt += str(index_value) + '\t' + '\t'.join(data[t, :]) + '\n'

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(txt)


class ObjectsModelOld(QtCore.QAbstractTableModel):
    """
    Class to populate a Qt table view with the properties of objects
    """

    def __init__(self,
                 objects: List[ALL_DEV_TYPES],
                 editable_headers: Dict[str, GCProp],
                 parent=None,
                 editable=False,
                 transposed=False,
                 check_unique: Union[None, List[str]] = None,
                 dictionary_of_lists: Union[None, Dict[str, List[ALL_DEV_TYPES]]] = None):
        """

        :param objects: list of objects associated to the editor
        :param editable_headers: Dictionary with the properties and the units and type {attribute: ('unit', type)}
        :param parent: Parent object: the QTableView object
        :param editable: Is the table editable?
        :param transposed: Display the table transposed?
        :param dictionary_of_lists: dictionary of lists for the Delegates
        """
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.parent = parent

        self.attributes = list(editable_headers.keys())

        self.attribute_types = [editable_headers[attr].tpe for attr in self.attributes]

        self.units = [editable_headers[attr].units for attr in self.attributes]

        self.tips = [editable_headers[attr].definition for attr in self.attributes]

        self.objects = objects

        self.editable = editable

        self.non_editable_attributes = [attr for attr in self.attributes if not editable_headers[attr].editable]

        self.check_unique = check_unique if check_unique is not None else list()

        self.r = len(self.objects)

        self.c = len(self.attributes)

        self.formatter = lambda x: "%.2f" % x

        self.transposed = transposed

        self.dictionary_of_lists = dictionary_of_lists if dictionary_of_lists is not None else dict()

        self.set_delegates()

    def set_delegates(self) -> None:
        """
        Set the cell editor types depending on the attribute_types array
        """

        if self.transposed:
            F = self.parent.setItemDelegateForRow
        else:
            F = self.parent.setItemDelegateForColumn

        for i in range(self.c):
            tpe = self.attribute_types[i]

            if tpe is bool:
                delegate = ComboDelegate(self.parent, [True, False], ['True', 'False'])
                F(i, delegate)

            elif tpe is str:

                if 'color' in self.attributes[i]:
                    delegate = ColorPickerDelegate(self.parent)
                else:
                    delegate = TextDelegate(self.parent)

                F(i, delegate)

            elif tpe is float:
                delegate = FloatDelegate(self.parent)
                F(i, delegate)

            elif tpe is complex:
                delegate = ComplexDelegate(self.parent)
                F(i, delegate)

            elif tpe is None:
                F(i, None)
                if len(self.non_editable_attributes) == 0:
                    self.non_editable_attributes.append(self.attributes[i])

            elif isinstance(tpe, EnumMeta):
                objects = list(tpe)
                values = [x.value for x in objects]
                delegate = ComboDelegate(self.parent, objects, values)
                F(i, delegate)

            elif tpe in [DeviceType.SubstationDevice,
                         DeviceType.AreaDevice,
                         DeviceType.ZoneDevice,
                         DeviceType.CountryDevice,
                         DeviceType.Technology,
                         DeviceType.ContingencyGroupDevice,
                         DeviceType.InvestmentsGroupDevice,
                         DeviceType.FuelDevice,
                         DeviceType.EmissionGasDevice,
                         DeviceType.GeneratorDevice]:

                objects = self.dictionary_of_lists[str(tpe.value)]
                values = [x.name for x in objects]
                delegate = ComboDelegate(self.parent, objects, values)
                F(i, delegate)

            else:
                F(i, None)

    def update(self):
        """
        update table
        """
        row = self.rowCount()
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        # whatever code
        self.endInsertRows()

    def flags(self, index):
        """
        Get the display mode
        :param index:
        :return:
        """
        if self.transposed:
            attr_idx = index.row()
        else:
            attr_idx = index.column()

        if self.editable and self.attributes[attr_idx] not in self.non_editable_attributes:
            return (
                    QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable)
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled

    def rowCount(self, parent=None):
        """
        Get number of rows
        :param parent:
        :return:
        """
        if self.transposed:
            return self.c
        else:
            return self.r

    def columnCount(self, parent=None):
        """
        Get number of columns
        :param parent:
        :return:
        """
        if self.transposed:
            return self.r
        else:
            return self.c

    def data_raw(self, r: int, c: int):
        """
        Get the data to display
        :param r: row index
        :param c: col index
        :return:
        """

        if self.transposed:
            obj_idx = c
            attr_idx = r
        else:
            obj_idx = r
            attr_idx = c

        attr = self.attributes[attr_idx]
        tpe = self.attribute_types[attr_idx]

        if tpe is Bus:
            return getattr(self.objects[obj_idx], attr).name
        else:
            return getattr(self.objects[obj_idx], attr)

    def data_with_type(self, index: QtCore.QModelIndex):
        """
        Get the data to display
        :param index:
        :return:
        """

        if self.transposed:
            obj_idx = index.column()
            attr_idx = index.row()
        else:
            obj_idx = index.row()
            attr_idx = index.column()

        attr = self.attributes[attr_idx]
        tpe = self.attribute_types[attr_idx]

        if tpe is Bus:
            return getattr(self.objects[obj_idx], attr).name
        else:
            return getattr(self.objects[obj_idx], attr)

    def data(self, index, role=None):
        """
        Get the data to display
        :param index:
        :param role:
        :return:
        """
        if index.isValid():
            if role == QtCore.Qt.ItemDataRole.DisplayRole:
                return str(self.data_with_type(index))
            elif role == QtCore.Qt.ItemDataRole.BackgroundRole:
                if 'color' in self.attributes[index.column()]:
                    return QtGui.QColor(str(self.data_with_type(index)))

        return None

    def setData(self, index, value, role=None):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        :return:
        """

        if self.transposed:
            obj_idx = index.column()
            attr_idx = index.row()
        else:
            obj_idx = index.row()
            attr_idx = index.column()

        tpe = self.attribute_types[attr_idx]

        # check taken values
        if self.attributes[attr_idx] in self.check_unique:
            taken = self.attr_taken(self.attributes[attr_idx], value)
        else:
            taken = False

        if not taken:
            if self.attributes[attr_idx] not in self.non_editable_attributes:

                if tpe is ContingencyGroup:
                    if value != "":
                        setattr(self.objects[obj_idx], self.attributes[attr_idx], ContingencyGroup(value))
                else:
                    setattr(self.objects[obj_idx], self.attributes[attr_idx], value)
            else:
                pass  # the column cannot be edited

        return True

    def attr_taken(self, attr, val):
        """
        Checks if the attribute value is taken
        :param attr:
        :param val:
        :return:
        """
        for obj in self.objects:
            if val == getattr(obj, attr):
                return True
        return False

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """
        Get the headers to display
        :param section:
        :param orientation:
        :param role:
        :return:
        """
        if role == QtCore.Qt.ItemDataRole.DisplayRole:

            if self.transposed:
                # for the properties in the schematic view
                if orientation == QtCore.Qt.Orientation.Horizontal:
                    return 'Value'
                elif orientation == QtCore.Qt.Orientation.Vertical:
                    if self.units[section] != '':
                        return self.attributes[section] + ' [' + self.units[section] + ']'
                    else:
                        return self.attributes[section]
            else:
                # Normal
                if orientation == QtCore.Qt.Orientation.Horizontal:
                    if self.units[section] != '':
                        return self.attributes[section] + ' [' + self.units[section] + ']'
                    else:
                        return self.attributes[section]
                elif orientation == QtCore.Qt.Orientation.Vertical:
                    return str(section) + ':' + str(self.objects[section])

        # add a tooltip
        if role == QtCore.Qt.ItemDataRole.ToolTipRole:
            if section < self.c:
                if self.units[section] != "":
                    unit = '\nUnits: ' + self.units[section]
                else:
                    unit = ''
                return self.attributes[section] + unit + ' \n' + self.tips[section]
            else:
                # somehow the index is out of range
                return ""

        return None

    def copy_to_column(self, index):
        """
        Copy the value pointed by the index to all the other cells in the column
        :param index: QModelIndex instance
        :return:
        """
        value = self.data_with_type(index=index)
        col = index.column()

        for row in range(self.rowCount()):

            if self.transposed:
                obj_idx = col
                attr_idx = row
            else:
                obj_idx = row
                attr_idx = col

            if self.attributes[attr_idx] not in self.non_editable_attributes:
                setattr(self.objects[obj_idx], self.attributes[attr_idx], value)
            else:
                pass  # the column cannot be edited

    def get_data(self):
        """

        :return:
        """
        nrows = self.rowCount()
        ncols = self.columnCount()
        data = np.empty((nrows, ncols), dtype=object)

        for j in range(ncols):
            for i in range(nrows):
                data[i, j] = self.data_raw(r=i, c=j)

        columns = [self.headerData(i, orientation=QtCore.Qt.Orientation.Horizontal,
                                   role=QtCore.Qt.ItemDataRole.DisplayRole) for i in range(ncols)]

        index = [self.headerData(i, orientation=QtCore.Qt.Orientation.Vertical,
                                 role=QtCore.Qt.ItemDataRole.DisplayRole) for i in range(nrows)]

        return index, columns, data

    def copy_to_clipboard(self):
        """

        :return:
        """
        if self.columnCount() > 0:

            index, columns, data = self.get_data()

            data = data.astype(str)

            # header first
            txt = '\t' + '\t'.join(columns) + '\n'

            # data
            for t, index_value in enumerate(index):
                txt += str(index_value) + '\t' + '\t'.join(data[t, :]) + '\n'

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(txt)
